from datetime import date, datetime
from typing import Any

from pypika import Query, functions
from pypika.enums import Dialects
from pypika.queries import QueryBuilder, Selectable
from pypika.terms import Case, CustomFunction, Field, Term

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    DateFormatMapping,
    SQLTranslator,
)


class ToMilliseconds(functions.Function):
    def __init__(self, field: Field):
        super().__init__("TO_MILLISECONDS", field)


class AthenaQueryBuilder(QueryBuilder):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=SQLDialect.ATHENA, **kwargs)

    # Does the same as the parent class, but with OFFSET before LIMIT
    def _apply_pagination(self, querystring: str) -> str:
        if self._offset:
            querystring += self._offset_sql()

        if self._limit is not None:
            querystring += self._limit_sql()

        return querystring


class AthenaQuery(Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> AthenaQueryBuilder:
        return AthenaQueryBuilder(**kwargs)


class AthenaTranslator(SQLTranslator):
    DIALECT = SQLDialect.ATHENA
    QUERY_CLS = AthenaQuery
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    SUPPORT_UNPIVOT = False
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE",
        integer="INTEGER",
        text="VARCHAR",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    DATE_FORMAT_MAPPING = DateFormatMapping(
        # https://prestodb.io/docs/current/functions/datetime.html#mysql-date-functions
        day_number="%d",
        month_number="%m",
        month_short="%b",
        month_full="%M",
        year="%Y",
    )
    REGEXP_OP = RegexOp.REGEXP_LIKE
    TO_DATE_OP = ToDateOp.DATE_PARSE
    FROM_DATE_OP = FromDateOp.DATE_FORMAT

    @classmethod
    def _add_date(cls, *, target_column: Field, duration: int, unit: str, dialect: Dialects | None = None) -> Term:
        # We need implement our own function for athena because Presto requires the units to be
        # quoted. PyPika's DateAdd function removes them by applying LiteralValue to the unit
        custom = CustomFunction("DATE_ADD", ["unit", "duration", "target"])
        return custom(Term.wrap_constant(unit.removesuffix("s")), duration, target_column)

    @staticmethod
    def _wrap_split_part(term: Term) -> Term:
        # Athena is the only backend for which SPLIT_PART returns NULL fox out-of-range
        # indexes. Also, IF_NULL is not available
        return Case().when(term.isnull(), "").else_(term)

    @classmethod
    def _interval_to_seconds(cls, value: Selectable) -> functions.Function:
        return ToMilliseconds(value) / 1000

    @classmethod
    def _cast_to_timestamp(cls, value: str | date | datetime | Field | Term) -> functions.Function:
        if isinstance(value, date):
            value = datetime(value.year, value.month, value.day)
        if isinstance(value, datetime):
            value = value.strftime("%Y-%m-%d %H:%M:%S%z")
        return super()._cast_to_timestamp(value)


SQLTranslator.register(AthenaTranslator)



"""
This file contain end-to-end tests for pipeline execution

TODO move to tests/backends/pandas_executor
"""

import json
from functools import partial
from os import path

import pandas as pd
import pytest
from pytest_mock import MockFixture

from tests.utils import assert_dataframes_equals
from weaverbird.backends.pandas_executor.pipeline_executor import (
    PipelineExecutionFailure,
    execute_pipeline,
    preview_pipeline,
)
from weaverbird.pipeline import Pipeline

df_domain_a = pd.read_csv(path.join(path.dirname(__file__), "fixtures/domain_a.csv"))
DOMAINS = {"domain_a": df_domain_a}


@pytest.fixture
def pipeline_executor():
    return partial(execute_pipeline, domain_retriever=lambda name: DOMAINS[name])


@pytest.fixture
def pipeline_previewer():
    return partial(preview_pipeline, domain_retriever=lambda name: DOMAINS[name])


def test_preview_pipeline(mocker: MockFixture, pipeline_previewer):
    df_to_json_spy = mocker.spy(pd.DataFrame, "to_json")
    result = json.loads(
        pipeline_previewer(
            Pipeline(
                steps=[
                    {"name": "domain", "domain": "domain_a"},
                ]
            )
        )
    )
    assert "data" in result
    assert len(result["data"]) == 3  # rows
    assert len(result["data"][0]) == 3  # columns
    assert result["schema"]["fields"] == [
        {"name": "colA", "type": "string"},
        {"name": "colB", "type": "integer"},
        {"name": "colC", "type": "integer"},
    ]
    assert result["offset"] == 0
    assert result["limit"] == 50
    assert result["total"] == 3

    # DataFrames must be exported with pandas' method to ensure NaN and dates are correctly converted
    df_to_json_spy.assert_called_once()


def test_preview_pipeline_limit(pipeline_previewer):
    result = pipeline_previewer(
        Pipeline(
            steps=[
                {"name": "domain", "domain": "domain_a"},
            ]
        ),
        limit=1,
    )
    assert json.loads(result)["data"] == [{"colA": "toto", "colB": 1, "colC": 100}]  # first row of the data frame


def test_preview_pipeline_limit_offset(pipeline_previewer):
    result = pipeline_previewer(
        Pipeline(
            steps=[
                {"name": "domain", "domain": "domain_a"},
            ]
        ),
        limit=3,
        offset=2,
    )
    assert json.loads(result)["data"] == [
        {"colA": "tata", "colB": 3, "colC": 25}  # third row of the data frame
        # no other row after that one
    ]


def test_extract_domain(pipeline_executor):
    df, _ = pipeline_executor(Pipeline(steps=[{"name": "domain", "domain": "domain_a"}]))

    assert_dataframes_equals(df, pd.DataFrame(df_domain_a))


def test_filter(pipeline_executor):
    df, _ = pipeline_executor(
        Pipeline(
            steps=[
                {"name": "domain", "domain": "domain_a"},
                {
                    "name": "filter",
                    "condition": {"column": "colA", "operator": "eq", "value": "tutu"},
                },
            ]
        )
    )

    assert_dataframes_equals(
        df,
        pd.DataFrame({"colA": ["tutu"], "colB": [2], "colC": [50]}),
    )


def test_rename(pipeline_executor):
    df, _ = pipeline_executor(
        Pipeline(
            steps=[
                {"name": "domain", "domain": "domain_a"},
                {"name": "rename", "toRename": [["colA", "col_a"], ["colB", "col_b"]]},
            ]
        )
    )

    assert_dataframes_equals(
        df,
        pd.DataFrame({"col_a": ["toto", "tutu", "tata"], "col_b": [1, 2, 3], "colC": [100, 50, 25]}),
    )


def test_errors(pipeline_executor):
    """
    It should provide helpful information when the pipeline execution fails, such as:
    - the step that encountered an error (nth and type)
    - the original exception message
    """
    with pytest.raises(PipelineExecutionFailure) as excinfo:
        pipeline_executor(
            Pipeline(
                steps=[
                    {"name": "domain", "domain": "domain_a"},
                    {
                        "name": "sort",
                        "columns": [{"column": "whatever", "order": "asc"}],
                    },
                ]
            )
        )
    exception_message = excinfo.value.message
    assert "Step #2" in exception_message
    assert "sort" in exception_message
    assert "whatever" in exception_message
    assert excinfo.value.details["index"] == 1
    assert excinfo.value.details["message"] == exception_message


def test_report(pipeline_executor):
    _, report = pipeline_executor(
        Pipeline(
            steps=[
                {"name": "domain", "domain": "domain_a"},
                {"name": "rename", "toRename": [["colA", "col_a"], ["colB", "col_b"]]},
            ]
        )
    )
    # there should be one step_report per step in the pipeline
    assert len(report.steps_reports) == 2



import pytest
from pandas import DataFrame

from weaverbird.backends.pandas_executor.steps.utils.cleaning import rename_duplicated_columns
from weaverbird.utils.iter import combinations


@pytest.mark.parametrize(
    ("columns", "expected_new_columns"),
    [
        (
            ["foo"],
            ["foo"],
        ),
        (
            ["foo", "foo"],
            ["foo", "foo_1"],
        ),
        (
            ["foo", "foo_1", "foo_2", "foo_3", "foo"],
            ["foo", "foo_1", "foo_2", "foo_3", "foo_4"],
        ),
        (
            ["foo_1", "foo_1"],
            ["foo_1", "foo_2"],
        ),
        (
            ["foo_1", "foo_1", "foo_2"],
            ["foo_1", "foo_3", "foo_2"],
        ),
        (
            ["foo", "foo", "foo_1"],
            ["foo", "foo_2", "foo_1"],
        ),
        (
            ["foo", "foo_3", "foo"],
            ["foo", "foo_3", "foo_4"],
        ),
        (
            ["foo", "foo", "foo_5", "foo_5"],
            ["foo", "foo_6", "foo_5", "foo_7"],
        ),
        (
            ["plop", "foo", "bar", "foo", "foo", "bar", "plip", "foo_2", "bar_1"],
            ["plop", "foo", "bar", "foo_3", "foo_4", "bar_2", "plip", "foo_2", "bar_1"],
        ),
    ],
)
def test_rename_duplicated_columns(columns, expected_new_columns):
    # ensure there are no duplicates in the output:
    assert len(expected_new_columns) == len(set(expected_new_columns))

    df = DataFrame(columns=columns)
    assert list(df.columns) == columns

    df_renamed = rename_duplicated_columns(df)
    assert list(df_renamed.columns) == expected_new_columns


def test_combinations():
    assert combinations(["A", "B", "C"]) == [
        ("A",),
        ("B",),
        ("C",),
        ("A", "B"),
        ("A", "C"),
        ("B", "C"),
        ("A", "B", "C"),
    ]



import logging

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline import Pipeline, steps
from weaverbird.pipeline.steps import DomainStep, JoinStep

logger = logging.getLogger(__name__)


_JOIN_VAR_KEY_NAME = "mongo_join_key"


def translate_join(step: JoinStep) -> list[MongoStep]:
    mongo_pipeline: list[dict] = []
    right = step.right_pipeline
    right_without_domain = Pipeline(steps=[])

    if isinstance(right, str):
        right_domain = DomainStep(name="domain", domain=right)
    else:
        try:
            assert isinstance(right, list)
        except AssertionError:  # in this case right is a Reference
            raise Exception(  # noqa: B904
                "References must be resolved before translating the pipeline"
            )
        if isinstance(right[0], DomainStep):
            right_domain = right[0]
            right_without_domain.steps = [s.copy(deep=True) for s in right[1:]]
        else:
            right_domain = DomainStep(**right[0])
            right_without_domain.steps = [getattr(steps, f"{s['name'].capitalize()}Step")(**s) for s in right[1:]]

    mongo_let: dict[str, str] = {}
    mongo_expr_and: list[dict[str, list[str]]] = []

    for idx, (left_on, right_on) in enumerate(step.on):
        var_name = _JOIN_VAR_KEY_NAME + f"_{idx}"
        mongo_let[var_name] = f"${left_on}"
        mongo_expr_and.append({"$eq": [f"${right_on}", f"$${var_name}"]})

    from weaverbird.backends.mongo_translator.mongo_pipeline_translator import translate_pipeline

    mongo_pipeline.append(
        {
            "$lookup": {
                "from": right_domain.domain,
                "let": mongo_let,
                "pipeline": translate_pipeline(right_without_domain)
                + [{"$match": {"$expr": {"$and": mongo_expr_and}}}],
                "as": "_vqbJoinKey",
            }
        }
    )

    if step.type == "inner":
        mongo_pipeline.append({"$unwind": "$_vqbJoinKey"})
    elif step.type == "left" or step.type == "left outer":
        mongo_pipeline.append({"$unwind": {"path": "$_vqbJoinKey", "preserveNullAndEmptyArrays": True}})
    else:
        mongo_pipeline.append({"$match": {"_vqbJoinKey": {"$eq": []}}})
        mongo_pipeline.append({"$unwind": {"path": "$_vqbJoinKey", "preserveNullAndEmptyArrays": True}})

    mongo_pipeline.append(
        {
            "$replaceRoot": {"newRoot": {"$mergeObjects": ["$_vqbJoinKey", "$$ROOT"]}},
        }
    )
    mongo_pipeline.append(
        {
            "$project": {"_vqbJoinKey": 0},
        },
    )
    return mongo_pipeline



from typing import Any


class WeaverbirdError(Exception):
    """Base class for weaverbird exceptions"""


class DuplicateError(WeaverbirdError):
    """Raised when an operation encountered duplicates where it was forbidden"""


class UnresolvedReferenceError(Exception):
    """Raised when a join step as a reference instead of a pipeline in input"""


class MissingColumnNameError(WeaverbirdError):
    """Raised when defining a substring but did not supply a new column name"""


class MissingTableNameError(WeaverbirdError):
    """Raised when defining a custom SQL query but didn't set a proper temporary table name"""


class PipelineFailure(WeaverbirdError):
    """Raised when an error happens on the pipeline"""

    def __init__(
        self,
        original_exception: Exception,
        step_name: str | None = None,
        step_config: dict[str, Any] | None = None,
        index: int | None = None,
    ):
        self.details: dict[str, Any] = {}
        if step_name and index:
            self.message = f"Step #{index + 1} ({step_name}) failed: {original_exception}"
            self.details["index"] = index
            self.details["step_config"] = step_config
        else:
            self.message = f"Internal failure: {original_exception}"
        self.details["message"] = self.message
        self.original_exception = original_exception
        super().__init__(self.message, self.details)


class PipelineTranslationFailure(PipelineFailure):
    """Raised when an error happens during the translation of the pipeline"""



import glob
import json
from copy import deepcopy
from datetime import datetime

import pytest
from jinja2.nativetypes import NativeEnvironment
from pydantic import BaseModel
from toucan_connectors.common import nosql_apply_parameters_to_query

from weaverbird.pipeline.conditions import ComparisonCondition, InclusionCondition
from weaverbird.pipeline.pipeline import (
    Pipeline,
    PipelineStep,
    PipelineStepWithVariables,
    PipelineWithVariables,
    remove_void_conditions_from_filter_steps,
    remove_void_conditions_from_mongo_steps,
)
from weaverbird.pipeline.steps import DomainStep, FilterStep, JoinStep, RollupStep, TrimStep
from weaverbird.pipeline.steps.aggregate import Aggregation
from weaverbird.pipeline.steps.append import AppendStep
from weaverbird.pipeline.steps.text import TextStep


class Case(BaseModel):
    filename: str
    data: dict
    context: dict
    expected_result: list


def jinja_renderer(query: dict | list[dict] | tuple | str, parameters: dict):
    if isinstance(query, dict):
        return {key: jinja_renderer(value, parameters) for key, value in deepcopy(query).items()}
    elif isinstance(query, list):
        rendered_query = [jinja_renderer(elt, parameters) for elt in deepcopy(query)]
        return rendered_query
    elif isinstance(query, tuple):
        return tuple(jinja_renderer(value, parameters) for value in deepcopy(query))
    elif isinstance(query, str):
        return NativeEnvironment().from_string(query).render(parameters)
    else:
        return query


def get_render_variables_test_cases():
    test_cases = []
    globs = glob.glob("./tests/fixtures/fixtures_templating/*.json")
    for file in globs:
        with open(file) as json_file:
            file_content = json.load(json_file)
            for test in file_content:
                case = Case(filename=file, data=test[0], context=test[1], expected_result=test[2])
                test_cases.append(case)
    return test_cases


cases = get_render_variables_test_cases()
ids = (x.filename for x in cases)


@pytest.mark.parametrize("case", cases, ids=ids)
def test_step_with_variables(case: Case):
    pipeline_with_variables = PipelineWithVariables(**case.data)

    pipeline = pipeline_with_variables.render(case.context, renderer=nosql_apply_parameters_to_query)

    expected_result = Pipeline(steps=case.expected_result)
    assert pipeline == expected_result


def test_custom_sql_step_with_variables():
    steps = [{"name": "customsql", "query": "{{ __front_var_0__ }}"}]
    variables = {"__front_var_0__": "-- DROP TABLE users;"}

    pipeline_with_variables = PipelineWithVariables(steps=steps)
    pipeline = pipeline_with_variables.render(variables, renderer=nosql_apply_parameters_to_query)

    # It should not have been rendered:
    assert pipeline.steps[0].query == "{{ __front_var_0__ }}"


def test_to_dict():
    pipeline = Pipeline(
        steps=[
            DomainStep(name="domain", domain="foobar"),
            RollupStep(
                name="rollup",
                hierarchy=["a", "b"],
                aggregations=[Aggregation(newcolumns=["a"], aggfunction="sum", columns=["a"])],
            ),
        ]
    )

    actual_dict = pipeline.dict()

    expected_dict = {
        "steps": [
            {"domain": "foobar", "name": "domain"},
            {
                "name": "rollup",
                "hierarchy": ["a", "b"],
                "aggregations": [{"new_columns": ["a"], "agg_function": "sum", "columns": ["a"]}],
            },
        ]
    }
    assert actual_dict == expected_dict
    assert pipeline == Pipeline(**pipeline.dict())


def test_skip_void_parameter_from_variables():
    void_step = {
        "name": "filter",
        "condition": {"column": "colA", "operator": "ge", "value": "{{ __front_var_0__ }}"},
    }

    simple_filter_step = {
        "name": "filter",
        "condition": {"column": "colB", "operator": "eq", "value": "{{ __front_var_1__ }}"},
    }

    composed_filter_step_and_ = {
        "name": "filter",
        "condition": {
            "and_": [
                {"column": "colC", "operator": "gt", "value": "{{ __front_var_2__ }}"},
                {"column": "ColD", "operator": "until", "value": datetime(2009, 1, 3, 0, 0)},
            ]
        },
    }

    composed_filter_step_or_and_ = {
        "name": "filter",
        "condition": {
            "or_": [
                {"column": "colE", "operator": "eq", "value": "{{ __front_var_3__ }}"},
                {"column": "colF", "operator": "lt", "value": "{{ __front_var_4__ }}"},
                {
                    "and_": [
                        {"column": "colG", "operator": "gt", "value": "{{ __front_var_5__ }}"},
                        {"column": "colH", "operator": "eq", "value": "{{ __front_var_6__ }}"},
                        {
                            "or_": [
                                {"column": "colI", "operator": "eq", "value": "toto"},
                                {
                                    "column": "colJ",
                                    "operator": "eq",
                                    "value": "{{ __front_var_7__ }}",
                                },
                                {
                                    "or_": [
                                        {"column": "colM", "operator": "gt", "value": "__VOID__"},
                                        {"column": "colN", "operator": "le", "value": "__VOID__"},
                                        {
                                            "and_": [
                                                {
                                                    "column": "colO",
                                                    "operator": "eq",
                                                    "value": "__VOID__",
                                                },
                                            ]
                                        },
                                        {
                                            "column": "colP",
                                            "operator": "in",
                                            "value": [
                                                "__VOID__",
                                                "{{ __front_var_11__ }}",
                                            ],
                                        },
                                        {
                                            "column": "colQ",
                                            "operator": "nin",
                                            "value": [False, "{{ __front_var_10__ }}"],
                                        },
                                    ]
                                },
                            ]
                        },
                    ]
                },
            ]
        },
    }
    composed_filter_step_and_all_voids = {
        "name": "filter",
        "condition": {
            "and_": [
                {"column": "colK", "operator": "gt", "value": "{{ __front_var_8__ }}"},
                {"column": "colL", "operator": "le", "value": "{{ __front_var_9__ }}"},
            ]
        },
    }

    steps = [
        {"domain": "foobar", "name": "domain"},
        void_step,
        composed_filter_step_or_and_,
        simple_filter_step,
        composed_filter_step_and_,
        composed_filter_step_and_all_voids,
    ]
    variables = {
        "__front_var_0__": "__VOID__",
        "__front_var_1__": 32,
        "__front_var_2__": "__VOID__",
        "__front_var_3__": "TEST TEST",
        "__front_var_4__": "__VOID__",
        "__front_var_5__": "__VOID__",
        "__front_var_6__": "VALUE",
        "__front_var_7__": "TOTO",
        "__front_var_8__": "__VOID__",
        "__front_var_9__": "__VOID__",
        "__front_var_10__": "__VOID__",
        "__front_var_11__": True,
    }

    pipeline_with_variables = PipelineWithVariables(steps=steps)
    pipeline = pipeline_with_variables.render(variables, renderer=jinja_renderer)
    steps = remove_void_conditions_from_filter_steps(pipeline.steps)
    # will render the with no __VOID__ in step, whatever the depth
    assert [s.dict() for s in steps] == [
        {"name": "domain", "domain": "foobar"},
        {
            "name": "filter",
            "condition": {
                "or_": [
                    {"column": "colE", "operator": "eq", "value": "TEST TEST"},
                    {
                        "and_": [
                            {"column": "colH", "operator": "eq", "value": "VALUE"},
                            {
                                "or_": [
                                    {"column": "colI", "operator": "eq", "value": "toto"},
                                    {"column": "colJ", "operator": "eq", "value": "TOTO"},
                                    {
                                        "or_": [
                                            {"column": "colP", "operator": "in", "value": [True]},
                                            {"column": "colQ", "operator": "nin", "value": [False]},
                                        ]
                                    },
                                ]
                            },
                        ]
                    },
                ]
            },
        },
        {"name": "filter", "condition": {"column": "colB", "operator": "eq", "value": 32}},
        {
            "name": "filter",
            "condition": {"and_": [{"column": "ColD", "operator": "until", "value": datetime(2009, 1, 3, 0, 0)}]},
        },
    ]


def test_skip_void_parameter_from_variables_for_mongo_steps():
    assert remove_void_conditions_from_mongo_steps(
        [{"$match": {"STORE_TYPE": {"$eq": "__VOID__"}, "DOUM": {"$eq": "wut"}}}]
    ) == [{"$match": {"DOUM": {"$eq": "wut"}}}]

    assert remove_void_conditions_from_mongo_steps(
        [
            {
                "$match": {
                    "$ne": None,
                    "$and": [
                        {"$or": [{"property2": "value2"}, {"property3": {}}]},
                        {"$nor": [{"property4": {}}, {"property5": {}}]},
                    ],
                    "$or": [{"x": {"$eq": "__VOID__"}}],
                    "$nor": [
                        {
                            "property9": {
                                "$and": [
                                    {
                                        "property7": {},
                                    }
                                ]
                            }
                        }
                    ],
                }
            }
        ]
    ) == [{"$match": {"$and": [{"$or": [{"property2": "value2"}]}], "$ne": None}}]

    assert remove_void_conditions_from_mongo_steps(
        [
            {
                "$match": {
                    "$eq": None,
                }
            },
            {"$group": {"_id": None, "_vqbPipelineInline": {"$push": "$$ROOT"}}},
            {
                "$lookup": {
                    "as": "_vqbPipelineToAppend_0",
                    "from": "slide_data-append",
                    "pipeline": [],
                }
            },
            {"$project": {"_vqbPipelinesUnion": {"$concatArrays": ["$_vqbPipelineInline", "$_vqbPipelineToAppend_0"]}}},
            {"$unwind": "$_vqbPipelinesUnion"},
            {"$replaceRoot": {"newRoot": "$_vqbPipelinesUnion"}},
            {"$project": {"_id": 0}},
        ]
    ) == [
        {"$match": {"$eq": None}},
        {"$group": {"_id": None, "_vqbPipelineInline": {"$push": "$$ROOT"}}},
        {"$lookup": {"as": "_vqbPipelineToAppend_0", "from": "slide_data-append", "pipeline": []}},
        {"$project": {"_vqbPipelinesUnion": {"$concatArrays": ["$_vqbPipelineInline", "$_vqbPipelineToAppend_0"]}}},
        {"$unwind": "$_vqbPipelinesUnion"},
        {"$replaceRoot": {"newRoot": "$_vqbPipelinesUnion"}},
        {"$project": {"_id": 0}},
    ]

    # to cover the rank query that have sometimes None values for prevRank
    assert remove_void_conditions_from_mongo_steps(
        [
            {"$match": {"domain": "cars"}},
            {"$sort": {"color": 1}},
            {"$group": {"_id": None, "_vqbArray": {"$push": "$$ROOT"}}},
            {
                "$project": {
                    "_vqbSortedArray": {
                        "$let": {
                            "vars": {
                                "reducedArrayInObj": {
                                    "$reduce": {
                                        "input": "$_vqbArray",
                                        "initialValue": {
                                            "a": [],
                                            "order": 0,
                                            "prevValue": None,
                                            "prevRank": None,
                                        },
                                        "in": {
                                            "$let": {
                                                "vars": {
                                                    "order": {
                                                        "$cond": [
                                                            {
                                                                "$ne": [
                                                                    "$$this.color",
                                                                    "$$value.prevValue",
                                                                ]
                                                            },
                                                            {"$add": ["$$value.order", 1]},
                                                            "$$value.order",
                                                        ]
                                                    },
                                                    "rank": {
                                                        "$cond": [
                                                            {
                                                                "$ne": [
                                                                    "$$this.color",
                                                                    "$$value.prevValue",
                                                                ]
                                                            },
                                                            {"$add": ["$$value.order", 1]},
                                                            "$$value.prevRank",
                                                        ]
                                                    },
                                                },
                                                "in": {
                                                    "a": {
                                                        "$concatArrays": [
                                                            "$$value.a",
                                                            [
                                                                {
                                                                    "$mergeObjects": [
                                                                        "$$this",
                                                                        {"color-rank": "$$rank"},
                                                                    ]
                                                                }
                                                            ],
                                                        ]
                                                    },
                                                    "order": "$$order",
                                                    "prevValue": "$$this.color",
                                                    "prevRank": "$$rank",
                                                },
                                            }
                                        },
                                    }
                                }
                            },
                            "in": "$$reducedArrayInObj.a",
                        }
                    }
                }
            },
            {"$unwind": "$_vqbSortedArray"},
            {"$replaceRoot": {"newRoot": "$_vqbSortedArray"}},
        ]
    ) == [
        {"$match": {"domain": "cars"}},
        {"$sort": {"color": 1}},
        {"$group": {"_id": None, "_vqbArray": {"$push": "$$ROOT"}}},
        {
            "$project": {
                "_vqbSortedArray": {
                    "$let": {
                        "in": "$$reducedArrayInObj.a",
                        "vars": {
                            "reducedArrayInObj": {
                                "$reduce": {
                                    "in": {
                                        "$let": {
                                            "in": {
                                                "a": {
                                                    "$concatArrays": [
                                                        "$$value.a",
                                                        [
                                                            {
                                                                "$mergeObjects": [
                                                                    "$$this",
                                                                    {"color-rank": "$$rank"},
                                                                ]
                                                            }
                                                        ],
                                                    ]
                                                },
                                                "order": "$$order",
                                                "prevRank": "$$rank",
                                                "prevValue": "$$this.color",
                                            },
                                            "vars": {
                                                "order": {
                                                    "$cond": [
                                                        {
                                                            "$ne": [
                                                                "$$this.color",
                                                                "$$value.prevValue",
                                                            ]
                                                        },
                                                        {"$add": ["$$value.order", 1]},
                                                        "$$value.order",
                                                    ]
                                                },
                                                "rank": {
                                                    "$cond": [
                                                        {
                                                            "$ne": [
                                                                "$$this.color",
                                                                "$$value.prevValue",
                                                            ]
                                                        },
                                                        {"$add": ["$$value.order", 1]},
                                                        "$$value.prevRank",
                                                    ]
                                                },
                                            },
                                        }
                                    },
                                    "initialValue": {
                                        "a": [],  # array based key field filtering in mongo should be kept
                                        "order": 0,
                                        "prevRank": None,
                                        "prevValue": None,
                                    },
                                    "input": "$_vqbArray",
                                }
                            }
                        },
                    }
                }
            }
        },
        {"$unwind": "$_vqbSortedArray"},
        {"$replaceRoot": {"newRoot": "$_vqbSortedArray"}},
    ]


@pytest.mark.parametrize(
    "step",
    (
        {
            "$project": {
                "Keyword": "$Keyword",
                "Diff_Organic": "$Diff_Organic",
                "_vqbToUnpivot": {
                    "$objectToArray": {
                        "Current_Month": {"$ifNull": ["$Current_Month", None]},
                        "Previous_Month": {"$ifNull": ["$Previous_Month", None]},
                    }
                },
            }
        },
        {"$group": {"_id": {}, "sum_value": {"$sum": "$Value1"}}},
    ),
)
def test_remove_void_conditions_from_mongo_step_should_only_apply_to_match_operator(step: dict):
    assert remove_void_conditions_from_mongo_steps(step) == step
    # A list of steps should start with a match-all operation
    assert remove_void_conditions_from_mongo_steps([step]) == [{"$match": {}}, step]


@pytest.mark.parametrize(
    "steps,expected_steps",
    [
        (
            [
                DomainStep(domain="domain__beers"),
                FilterStep(condition=ComparisonCondition(column="beer_kind", operator="eq", value="__VOID__")),
                JoinStep(
                    type="left",
                    on=[("name", "name")],
                    right_pipeline=[
                        DomainStep(domain="domain__beers"),
                        TrimStep(columns=["name"]),
                        FilterStep(condition=ComparisonCondition(column="beer_kind", operator="eq", value="__VOID__")),
                        TextStep(name="text", text="var_two", new_column="var_two"),
                    ],
                ),
                AppendStep(
                    pipelines=[
                        [
                            FilterStep(
                                condition=ComparisonCondition(column="beer_kind", operator="eq", value="__VOID__")
                            ),
                        ],
                        "some-domain",
                        [
                            DomainStep(domain="domain__beers"),
                            TrimStep(columns=["name"]),
                            FilterStep(
                                condition=ComparisonCondition(column="beer_kind", operator="eq", value="__VOID__")
                            ),
                        ],
                        "__VOID__",
                    ]
                ),
            ],
            [
                # First filter step should have been removed
                DomainStep(domain="domain__beers"),
                JoinStep(
                    type="left",
                    on=[("name", "name")],
                    right_pipeline=[
                        # First filter step should have been removed here too
                        DomainStep(domain="domain__beers"),
                        TrimStep(columns=["name"]),
                        TextStep(name="text", text="var_two", new_column="var_two"),
                    ],
                ),
                AppendStep(
                    pipelines=[
                        # First pipeline should have been removed
                        # regular domains should be left untouched
                        "some-domain",
                        [
                            DomainStep(domain="domain__beers"),
                            TrimStep(columns=["name"]),
                            # filter step should have been removed from here too
                        ],
                        # regular domains should be left untouched
                        "__VOID__",
                    ]
                ),
            ],
        ),
        (
            [
                JoinStep(
                    type="left",
                    on=[("name", "name")],
                    right_pipeline=[
                        FilterStep(condition=ComparisonCondition(column="beer_kind", operator="eq", value="__VOID__"))
                    ],
                )
            ],
            # The join step should be completely removed
            [],
        ),
        (
            [JoinStep(type="left", on=[("name", "name")], right_pipeline="__VOID__")],
            # The join step should be left untouched
            [JoinStep(type="left", on=[("name", "name")], right_pipeline="__VOID__")],
        ),
        (
            [
                AppendStep(
                    pipelines=[
                        [FilterStep(condition=ComparisonCondition(column="beer_kind", operator="eq", value="__VOID__"))]
                    ]
                )
            ],
            # the append step should have been completely removed
            [],
        ),
    ],
)
def test_remove_void_conditions_from_filter_steps_with_combinations(
    steps: list[PipelineStep | PipelineStepWithVariables],
    expected_steps: list[PipelineStep | PipelineStepWithVariables],
) -> None:
    assert remove_void_conditions_from_filter_steps(steps) == expected_steps


def test_pipeline_with_refs_variables_and_date_validity():
    PipelineWithVariables(
        steps=[
            {"domain": {"type": "ref", "uid": "83ff1fa2-d186-4a7b-a53b-47c901a076c7"}, "name": "domain"},
            {
                "name": "append",
                "pipelines": [
                    {"type": "ref", "uid": "3ecb05aa-1312-4797-bf9a-65f736bb2993"},
                    [
                        {"name": "domain", "domain": "foo"},
                        {
                            "name": "filter",
                            "condition": {
                                "column": "date",
                                "operator": "from",
                                "value": {
                                    "date": "{{TODAY}}",
                                    "duration": "year",
                                    "operator": "until",
                                    "quantity": 1,
                                },
                            },
                        },
                    ],
                ],
            },
        ]
    )


@pytest.mark.parametrize(
    "steps,expected_steps",
    [
        (
            [
                DomainStep(domain="domain__beers"),
                FilterStep(condition=InclusionCondition(column="beer_kind", operator="in", value="__VOID__")),
                TextStep(name="text", text="var_two", new_column="var_two"),
                FilterStep(condition=InclusionCondition(column="beer_kind", operator="in", value=[])),
                FilterStep(condition=InclusionCondition(column="beer_kind", operator="in", value=["__VOID__"])),
                FilterStep(condition=InclusionCondition(column="beer_kind", operator="in", value=["__VOID__", "colA"])),
            ],
            [
                DomainStep(domain="domain__beers"),
                TextStep(name="text", text="var_two", new_column="var_two"),
                FilterStep(condition=InclusionCondition(column="beer_kind", operator="in", value=[])),
                FilterStep(condition=InclusionCondition(column="beer_kind", operator="in", value=[])),
                FilterStep(condition=InclusionCondition(column="beer_kind", operator="in", value=["colA"])),
            ],
        )
    ],
)
def test_keep_condition_if_empty_list_from_filter_steps(
    steps: list[PipelineStep | PipelineStepWithVariables],
    expected_steps: list[PipelineStep | PipelineStepWithVariables],
):
    assert remove_void_conditions_from_filter_steps(steps) == expected_steps



from numpy import nan as nan_value
from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import PivotStep

PIVOT_NULL_VALUE = "__WEAVERBIRD_PIVOT_NULL_VALUE__"


def execute_pivot(
    step: PivotStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    # Create alias for null values in indexed columns before running pivot to avoid removing null values
    for idx_column in step.index:
        df[idx_column] = df[idx_column].fillna(PIVOT_NULL_VALUE)
    df[PIVOT_NULL_VALUE] = nan_value

    pivoted_df = df.pivot_table(
        values=step.value_column,
        index=step.index,
        columns=step.column_to_pivot,
        aggfunc="mean" if step.agg_function == "avg" else step.agg_function,
    ).reset_index()

    # Replace alias with null value
    for idx_column in step.index:
        pivoted_df[idx_column] = pivoted_df[idx_column].replace(PIVOT_NULL_VALUE, nan_value)

    pivoted_df.columns.name = None
    return pivoted_df



from typing import Any

from pandas import Series, to_datetime, to_numeric
from pandas import __version__ as pd_version
from pandas import concat as pd_concat


def cast_to_float(s: Series) -> Series:
    return to_numeric(s, errors="coerce")  # cast errors will result in NaN values


def cast_to_int(s: Series) -> Series:
    s_float = cast_to_float(s)
    # Since .astype(int) would fail on NaN values, treat them separately:
    s_integer = s_float[s_float.notna()].astype(int)
    s_nan = s_float[s_float.isna()]
    # Pandas operate on NanoSecond Timestamp but we want to output MilliSecond Timestamps
    if str(s.dtype).startswith("datetime64"):
        s_integer = s_integer // 10**6
    return pd_concat([s_integer, s_nan]).sort_index()


def cast_to_str(s: Series) -> Series:
    return s.astype(str)


def cast_to_datetime(s: Series) -> Series:
    is_casting_timestamps = str(s.dtype).startswith("int")

    extra_kwargs: dict[str, Any]
    # in pandas 2, "format" can be "mixed", which means that pandas should figure out every format
    # individually. In pandas 1, this is the default, and "mixed" is interpredted as an strftime string
    if pd_version.startswith("2"):
        # ValueError('cannot specify both format and unit')
        extra_kwargs = {"unit": "ms"} if is_casting_timestamps else {"format": "mixed", "unit": None}
    else:
        extra_kwargs = {"unit": "ms" if is_casting_timestamps else None}
    return to_datetime(
        s,
        errors="coerce",  # cast errors will result in NaT values
        **extra_kwargs,
    )


def cast_to_bool(s: Series) -> Series:
    return s.astype(bool)
